

from abc import ABC, abstractmethod
from typing import Union
from bergen.utils import ExpansionError, expandInputs, shrinkOutputs
from bergen.messages.assignation import AssignationMessage
from bergen.schema import Template
from bergen.constants import ACCEPT_GQL, OFFER_GQL, SERVE_GQL
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node, Pod
import inspect
import sys
from aiostream import stream
import concurrent
from functools import partial
logger = logging.getLogger()

import json


class BaseHelper(ABC):

    def __init__(self, peasent) -> None:
        self.peasent = peasent
        pass

    @abstractmethod
    async def pass_yield(self, message, value):
        pass

    @abstractmethod
    async def pass_progress(self, message, value):
        pass

    @abstractmethod
    async def pass_result(self, message, value):
        pass

    @abstractmethod
    async def pass_exception(self, message, exception):
        pass


class AssignationHelper():

    def __init__(self, peasent_helper: BaseHelper, assignation) -> None:
        self.peasent_helper = peasent_helper
        self.assignation = assignation
        pass


    def progress(self, value):
        self.peasent_helper.pass_progress(self.message, value)


async def shrink(node_template_pod, outputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await shrinkOutputs(node=node_template_pod, outputs=outputs)

    if isinstance(node_template_pod, Template):
        kwargs = await shrinkOutputs(node=node_template_pod.node, outputs=outputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await shrinkOutputs(node=node_template_pod.template.node, outputs=outputs)


    return kwargs



async def expand(node_template_pod, inputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await expandInputs(node=node_template_pod, inputs=inputs)

    if isinstance(node_template_pod, Template):
        kwargs = await expandInputs(node=node_template_pod.node, inputs=inputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await expandInputs(node=node_template_pod.template.node, inputs=inputs) 
    return kwargs



threadhelper = None
try: 
    threadhelper = asyncio.to_thread
except:
    logger.warn("Threading does not work below Python 3.9")

    
class BasePeasent(ABC):
    ''' Is a mixin for Our Bergen '''
    helperClass = None


    def __init__(self, *args, loop = None, name = None, **kwargs) -> None:
        
        
        self.unique_name = name or namegenerator.gen()
        self.loop = loop or asyncio.get_event_loop()
        self.pods = {}


        self.nodes_to_register = [] # for Nodes to register
        self.template_function_set = [] # for Templates to register
        self.pod_function_set = [] # For Pods to register

        self.peasent_helper = self.helperClass(self)
    

        self.podid_function_map = {}
        self.podid_pod_map = {}
        super().__init__(*args, loop=loop, **kwargs)


    def register(self, node_template_pod: Union[Node, Template], **kwargs):

        threaded = kwargs.get("threaded", False)
        expanded = kwargs.get("threaded", True)


        def real_decorator(function):
            is_coroutine = inspect.iscoroutinefunction(function)
            is_asyncgen = inspect.isasyncgenfunction(function)
            is_function = inspect.isfunction(function)


            async def wrapper(message: AssignationMessage):
                
                assignation_helper = AssignationHelper(self.peasent_helper, message)
                kwargs = await expand(node_template_pod,message.data.inputs)

                if is_coroutine:
                    result = await function(assignation_helper, **kwargs)
                    outputs = await shrink(node_template_pod, result)
                    await self.peasent_helper.pass_result(message, outputs)

                elif is_asyncgen:
                    yieldstream = stream.iterate(function(assignation_helper, **kwargs))
                    lastresult = None
                    async with yieldstream.stream() as streamer:
                        async for result in streamer:
                            lastresult = await shrink(node_template_pod, result)
                            await self.peasent_helper.pass_yield(message, lastresult)

                    await self.peasent_helper.pass_result(message, lastresult)

                elif is_function:  
                    if threadhelper:
                        result = await threadhelper(function, assignation_helper, **kwargs)
                    else:
                        result = function(assignation_helper, **kwargs)
                        #helper.end(result)
                    #result = await self.loop.run_in_executor(self.executor, partial(function,**kwargs), assignation_helper)
                    outputs = await shrink(node_template_pod, result)
                    await self.peasent_helper.pass_result(message, outputs)
                
                else:
                    raise NotImplementedError("We do not allow for non async functions right now")
                
            
            wrapper.__name__ = function.__name__

            if isinstance(node_template_pod, Node):
                logger.info(f"Registering Template for {node_template_pod.name}")
                self.nodes_to_register.append({
                    "node": node_template_pod,
                    "params": kwargs,
                    "function": wrapper

                })

            if isinstance(node_template_pod, Template):
                logger.info(f"Already existing Template that we are providing for {node_template_pod.id}")
                self.template_function_set.append((node_template_pod,wrapper))


            if isinstance(node_template_pod, Pod):
                logger.info(f"Already existing Template that we are providing for {node_template_pod.id}")
                self.pod_function_set.append((node_template_pod,wrapper))



            return wrapper


        return real_decorator



    async def setup_and_run(self):

        logger.info("Offering Our Services")
        self.peasent = await SERVE_GQL.run_async(ward=self.main_ward, variables = {"name":self.unique_name})
        logger.info("Our Offer was Accesspted")


        async def offer(node: Node, params: dict, function):
            peasent_template = await OFFER_GQL.run_async(ward=self.main_ward, variables= {"node": node.id, "params": params, "peasent": self.peasent.id})
            return (peasent_template, function)

        logger.info("We are registering our functions as potential Templates")
        template_function_set = await asyncio.gather(*[offer(**values) for values in self.nodes_to_register])

        # We can now combing reigstered templates and already registered templates
        all_templates_function_set =  template_function_set + self.template_function_set

        logger.info("We are our Pods")
        async def accept(template: Template, function):
            peasent_pod = await ACCEPT_GQL.run_async(ward=self.main_ward, variables= {"template": template.id, "peasent": self.peasent.id})
            return (peasent_pod, function)


        pod_function_set = await asyncio.gather(*[accept(*values) for values in all_templates_function_set])
       
        # We can now combing potential templates pods and already potential pods
        all_pod_function_set = pod_function_set + self.pod_function_set

        self.podid_function_map = {value[0].id: value[1] for value in all_pod_function_set}
        self.podid_pod_map = {value[0].id: value[0] for value in all_pod_function_set}


        logger.info(f"Following functions have been allowed! {[value[1].__name__ for value in all_pod_function_set]}")
        for value in all_pod_function_set:
            logger.info(f"Requesting to host {value[1].__name__} as Pod {value[0].id}" )


        logger.debug("Configuring")
        await self.configure()



    @abstractmethod
    async def configure(self) -> str:
        raise NotImplementedError("Please overwrite")

    async def run_async(self):
        await self.setup_and_run()

    def run(self):
        if self.loop.is_running():
            logger.error("Cannot do this, please await run_asnyc()")
        else:
            task = self.loop.create_task(self.setup_and_run())

            async def _async(task):
                return await asyncio.wait([task])

            self.loop.run_until_complete(_async(task))

        # we enter a never-ending loop that waits for data
        # and runs callbacks whenever necessary.
        





from bergen.schema import Assignation, Node, Peasent, PeasentTemplate, Pod, Provision, Transcript, VartPod, Volunteer
from bergen.query import TypedGQL


NEGOTIATION_GQL = TypedGQL("""

  mutation Negotiate($clientType: ClientType!, $local: Boolean!) {
  negotiate(clientType: $clientType) {
    array {
      path
      params
    }
    timestamp
    user {
        username
    }
    extensions
    models {
        identifier
        point {
            name
            host (local: $local)
            port
            type
        }
    }
    postman {
        type
        kwargs
    }
    points {
      host
      name
      models {
        identifier
        extenders
      }
    }
    
  }
  }
""", Transcript)

 
# Peasent Constants

SERVE_GQL = TypedGQL("""
    mutation Serve($name: String!){
        serve(name:$name){
            id
            name
            
        }
    }
""", Peasent)




OFFER_GQL = TypedGQL("""
    mutation Offer($node: ID!, $params: GenericScalar!, $peasent: ID!){
        offer(node: $node, params: $params, peasent: $peasent ){
            id
            name 
        }
    }
""", PeasentTemplate)


ACCEPT_GQL = TypedGQL("""
    mutation AcceptMutation($peasent: ID!, $template: ID!){
        accept(peasent: $peasent, template: $template){
            id
        }
        }
""", Pod)











VOLUNTEER_GQL = TypedGQL("""
        mutation Volunteer($nodeid: ID!, $name: String!, $version: String!){
            volunteer(node: $nodeid, name: $name, version: $version){
                node {
                    id
                    name
                    package
                }
                identifier
                id
            }
        }
    """, Volunteer)



MARK_GQL = TypedGQL("""
        mutation Mark($message: String!, $assignation: ID!, $level: AssignationStatus) {
            mark(message: $message, assignation: $assignation, level: $level){
                registered
            }
        }
""", dict)


END_GQL = TypedGQL("""
        mutation End($outputs: GenericScalar!, $assignation: ID!) {
            end(outputs: $outputs, assignation: $assignation){
                registered
            }
        }
""", dict)

YIELD_GQL = TypedGQL("""
        mutation Yield($outputs: GenericScalar!, $assignation: ID!) {
            yield(outputs: $outputs, assignation: $assignation){
                registered
            }
        }
""", dict)


QUEUE_GQL = TypedGQL("""
    subscription Queue($id: ID!) {
        queue(volunteer: $id){
            id
            status
            volunteer {
                id
                node {
                    id
                    name
                }
                identifier
            }
        }
    }
""", VartPod)


HOST_GQL = TypedGQL("""
                        subscription Host($pod: ID!) {
                host(pod: $pod){
                    pod {
                        id
                    }
                    id
                    inputs
                }
                }
""", Assignation)


PROVIDE_GQL = TypedGQL("""
        subscription Provide($reference: String!, $node: ID!, $selector: SelectorInput!){
            provide(node: $node , selector: $selector, reference: $reference){
                pod {
                    id
                    status
                }
                node {
                    name
                }
                reference
                status
                statusmessage

            }
        }      
    """, Provision)



ASSIGN_GQL = TypedGQL("""
        subscription Assignation($inputs: GenericScalar!, $pod: ID!, $reference: String!) {
            assign(inputs: $inputs, pod: $pod, reference: $reference){
                inputs
                outputs
                status
                statusmessage
            }
        }     
""", Assignation)





from bergen.registries.arnheim import get_current_arnheim


class BaseExtender:
    extension = None

    def getSettings(self):
        return get_current_arnheim().getExtensionSettings(self.extension)



    pass


from bergen.types.model import ArnheimModel
from bergen.registries.matcher import get_current_matcher
from bergen.enums import TYPENAMES
from bergen.schema import Node
import logging

logger = logging.getLogger(__name__)

class ExpansionError(Exception):
    pass



async def expandInputs(node: Node, inputs: dict) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    kwargs = {}
    for port in node.inputs:
        if port.key not in inputs:
            if port.required:
                raise ExpansionError(f"We couldn't expand {port.key} because it wasn't provided by our Inputs, wrong assignation!!!")
            else:
                break

        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
            instance =  await modelClass.asyncs.get(id=inputs[port.key])
            kwargs[port.key] = instance
        else:
            kwargs[port.key] = inputs[port.key]

    return kwargs



async def shrinkOutputs(node: Node, outputs: dict) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    kwargs = {}
    for port in node.outputs:
        if port.key not in outputs:
            break

        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            instance = outputs[port.key]
            if isinstance(instance, ArnheimModel):
                kwargs[port.key] = outputs[port.key].id
            else:
                kwargs[port.key] = outputs[port.key]
        else:
            kwargs[port.key] = outputs[port.key]

    return kwargs



async def shrinkInputs(node: Node, inputs: dict) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    kwargs = {}
    for port in node.inputs:
        if port.key not in inputs:
            break

        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            instance = inputs[port.key]
            if isinstance(instance, ArnheimModel):
                kwargs[port.key] = inputs[port.key].id
            else:
                kwargs[port.key] = inputs[port.key]
        else:
            kwargs[port.key] = inputs[port.key]

    return kwargs


async def expandOutputs(node: Node, outputs: dict, strict=False) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    kwargs = {}
    for port in node.outputs:
        if port.key not in outputs:
            break

        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            try:
                modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
                instance =  await modelClass.asyncs.get(id=outputs[port.key])
                kwargs[port.key] = instance
            except AssertionError as e:
                if strict: raise e
                logger.error(f"Couldn't expand outputs make sure to import the schema for {port.identifier}")
                kwargs[port.key] = outputs[port.key]
           
        else:
            kwargs[port.key] = outputs[port.key]

    return kwargs










class UserPrettifier(object):
    ''' Just a little helper to make our User look nicer in Jupyter Notebooks'''

    def _repr_html_(self):
        string = f"{self.username}</br>"
        if self.avatar is not None:
            string += f"<img src='{self.avatar.avatar}' alt='userimage'></img>"
        
        return string


import re
from typing import Generator, Generic, Type, TypeVar



gqlparsed_with_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)\((?P<arguments>[^\)]*)\)[\s]*{[\s]*(?P<firstchild>[^\(:]*).*")
gqlparser_without_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)[\s]*{[\s]*(?P<firstchild>[^\(\{\s:]*).*")

class GQLException(Exception):
    pass


MyType = TypeVar("MyType")


class GQL(object):

    def __init__(self, query: str) -> None:
        self.query = query
        self.variables = None
        self._type = None
        self.m = gqlparsed_with_variables.match(self.query)
        self.has_variables = True
        if not self.m:
            self.m = gqlparser_without_variables.match(self.query)
            self.has_variables = False
            if not self.m:
                raise GQLException("Illformed request")

    def combine(self, variables: dict):
        self.variables = variables
        return {"query": self.query, "variables": self.variables}
        
    def parsed(self):
        assert self.variables and self.query, "Please specify query and set variables before parsing"
        return {"query": self.query, "variables": self.variables}

    @property
    def firstchild(self):
        return self.m.group("firstchild")

    @property
    def operation_name(self):
        return self.m.group("operation")

    @property
    def type(self):
        return self.m.group("type")

    def extract(self, result: dict):
        return result[self.firstchild]



class QueryList(list):


     def _repr_html_(self) -> str:
        listified = list(self)
        string = "Result List with <br/><table>"
        for x, item in enumerate(listified):
             string += f"""
                <tr>
                    <td>{x}<td/>
                    <td>{item._repr_list_() if hasattr(item,"_repr_list_") else str(item)}</td>
                </tr>

            """
        string + "</table>"
        return string


    
class TypedGQL(GQL, Generic[MyType]):

    def __init__(self, query: str, cls: Type[MyType], aslist=False):
        self.aslist = aslist
        self.query = query
        self._cls = cls
        super().__init__(query)

    @property
    def cls(self) -> MyType:
        return self._cls


    def run(self, ward=None, variables=None, **kwargs) -> MyType:
        from bergen.registries.arnheim import get_current_arnheim
        ward = ward or self._cls.get_ward()
        returnedobject = ward.run(self, variables=variables, **kwargs)
        assert returnedobject is not None, "We received nothing back from the Server! Refine your Query!"
        if isinstance(returnedobject,list): return [self.cls(**item) for item in returnedobject]
        return self.cls(**returnedobject)

    
    async def run_async(self, ward=None, variables=None, **kwargs) -> MyType:
        from bergen.registries.arnheim import get_current_arnheim
        ward = ward or self._cls.get_ward()
        returnedobject = await ward.run_async(self, variables=variables,**kwargs)
        
        assert returnedobject is not None, "We received nothing back from the Server! Refine your Query!"
        if isinstance(returnedobject,list): return [self.cls(**item) for item in returnedobject]
        return self.cls(**returnedobject)


    def subscribe(self, ward=None, **kwargs) -> Generator[MyType, None,None]:
        from bergen.registries.arnheim import get_current_arnheim
        ward = ward or get_current_arnheim().getWard()
        return ward.subscribe(self, **kwargs)

    def _repr_html_(self) -> str:
        string = "Query <br/><table>"
        string + "</table>"
        return string


def DelayedGQL(gqlstring):
    return lambda model : TypedGQL(gqlstring, model)



class BaseWidget:

  def __init__(self, type) -> None:
      assert type is not None, "Please provide a Valid Type for the Widget"
      self.type = type
      super().__init__()


  def serialize(self):
      return {
          "type": type

      }


  def deserialize(self, registry):
      raise NotImplementedError("Please specifiy this")



from bergen.messages.types import PROVISION
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import Optional


class ProvisionParams(BaseModel):
    pass

class ProvisionMetaExtensionsModel(MessageMetaExtensionsModel):
    progress: Optional[str]
    callback: Optional[str]

class ProvisionMetaModel(MessageMetaModel):
    type: str = PROVISION
    extensions: Optional[ProvisionMetaExtensionsModel]

class ProvisionDataModel(MessageDataModel):
    
    #Meta
    parent: Optional[int]
    id: int
    reference: str

    #Inputs
    node: Optional[int] #TODO: Maybe not optional
    pod: Optional[int]
    template: Optional[int]
    params: Optional[ProvisionParams]
    
    #Status
    status: Optional[str]
    statusmessage: Optional[str]

class ProvisionMessage(MessageModel):
    data: ProvisionDataModel
    meta: ProvisionMetaModel


from bergen.extenders.pod import PodExtender
from bergen.types.model import ArnheimModelManager
from bergen.extenders.user import UserExtender
from bergen.types.node.inputs import Inputs, Outputs
from bergen.queries.delayed.node import CREATE_NODE_MUTATION, NODE_FILTER_QUERY, NODE_QUERY
from bergen.queries.delayed.pod import POD_QUERY
from bergen.extenders.node import NodeExtender
from bergen.schema import Node as SchemaNode
from bergen.schema import Pod as SchemaPod
from bergen.schema import User as SchemaUser
from bergen.schema import *
try:
	# python 3.8
	from typing import ForwardRef, Type
except ImportError:
	# ForwardRef is private in python 3.6 and 3.7
	from typing import _ForwardRef as ForwardRef, Type

Node = ForwardRef('Node')

class NodeManager(ArnheimModelManager[Node]):

    def get_or_create(self, inputs: Type[Inputs] = None, outputs: Type[Outputs] = None , **kwargs) -> Node:
        
        parsed_inputs = inputs.serialized
        parsed_outputs = outputs.serialized
        
        node = CREATE_NODE_MUTATION(self.model).run(variables={
            "inputs" : parsed_inputs,
            "outputs": parsed_outputs,
            **kwargs

        })
        return node


class Node(NodeExtender, SchemaNode):
    __slots__ = ("_loop", "_force_sync", "_postman")

    objects = NodeManager()

    class Meta:
        overwrite_default = True
        identifier = "node"
        filter = NODE_FILTER_QUERY
        get = NODE_QUERY



class Pod(PodExtender, SchemaPod):
    __slots__ = ("_loop", "_force_sync", "_postman")

    objects = NodeManager()

    class Meta:
        overwrite_default = True
        identifier = "pod"
        get = POD_QUERY


class User(UserExtender, SchemaUser):

    class Meta:
        overwrite_default = True
        identifier = "user"



from abc import abstractmethod
from urllib.parse import non_hierarchical

from oauthlib.oauth2.rfc6749.clients.mobile_application import MobileApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session
from bergen.auths.base import AuthError, BaseAuthBackend
from bergen.enums import ClientType
import logging

logger = logging.getLogger(__name__)

try:
    from bergen.auths.implicit.widgets.login import LoginDialog
    # this has to be here because QtWebENgineWIdgets must be imported before a QCore Application
    Dialog = LoginDialog

except Exception as e:
    logger.debug(e)
    Dialog = None


class ImplicitApplication(BaseAuthBackend):


    def __init__(self, client_id = None, redirect_uri = "http://localhost:3000/callback", host="localhost", port= 8000, protocol = "http", scopes= ["read"], parent=None, **kwargs) -> None:
        self.client_id = client_id
        assert self.client_id is not None, "Please provide a client_id argument"
        # TESTED, just redirecting to Google works in normal browsers
        # the token string appears in the url of the address bar
        self.redirect_uri = redirect_uri

        # Generate correct URLs
        self.base_url = f"{protocol}://{host}:{port}/o/"
        self.auth_url = self.base_url + "authorize"
        self.token_url = self.base_url + "token"

        # If you want to have a hosting QtWidget
        self.parent = parent

        self.token = None


        self.mobile_app_client = MobileApplicationClient(client_id)

        # Create an OAuth2 session for the OSF
        self.session = OAuth2Session(
            client_id, 
            self.mobile_app_client,
            scope=" ".join(scopes), 
            redirect_uri=self.redirect_uri,
        )


        super().__init__()

    def getToken(self, loop=None) -> str:

        try:
            if not self.token:
                token, result = Dialog.getToken(backend=self, parent=self.parent)
                if result:
                    self.token = token
                else:
                    raise AuthError("Couldn't return a proper token")

        except Exception as e:
            raise NotImplementedError("It appears that you have not Installed PyQt5")
        
        return self.token["access_token"] # We actually get a fully fledged thing back


    def getClientType(self):
        return ClientType.EXTERNAL

    def getProtocol(self):
        return "http"



from bergen.types.model import ArnheimModel


def parse_kwargs(kwargs):

    parsed_kwargs = {}

    for key, value in kwargs.items():
        if isinstance(value, ArnheimModel):
            assert value.id is not None, "Cannot parse a ArnheimModel as Argument if it doesn't provide an ID"
            parsed_kwargs[key] = value.id
        else:
            parsed_kwargs[key] = value


    return parsed_kwargs




from abc import ABC, abstractmethod
from re import template
from bergen.messages.allowance import AllowanceMessage
from typing import Union
from bergen.utils import ExpansionError, expandInputs
from bergen.messages.assignation import AssignationMessage
from bergen.schema import AssignationStatus, Template
from bergen.constants import OFFER_GQL, SERVE_GQL
from bergen.peasent.base import BaseHelper, BasePeasent
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node
import inspect
import sys

logger = logging.getLogger()


class WebsocketHelper(BaseHelper):

    async def pass_yield(self, message, value):
        message.data.outputs = value
        message.data.status = AssignationStatus.YIELD
        await self.peasent.send_to_connection(message)

    async def pass_progress(self, message, value):
        message.data.status = AssignationStatus.PROGRESS#
        message.data.statusmessage = str(value)
        await self.peasent.send_to_connection(message)
        pass

    async def pass_result(self,message, value):
        message.data.outputs = value
        message.data.status = AssignationStatus.DONE
        await self.peasent.send_to_connection(message)

    async def pass_exception(self,message, exception):
        message.data.status = AssignationStatus.ERROR#
        message.data.statusmessage = str(exception)
        await self.peasent.send_to_connection(message)
        pass

class WebsocketPeasent(BasePeasent):
    helperClass = WebsocketHelper
    ''' Is a mixin for Our Bergen '''

    def __init__(self, *args, host= None, port= None, ssl = False,**kwargs) -> None:
        self.websocket_host = host
        self.websocket_port = port
        self.websocket_protocol = "wss" if ssl else "ws"

        self.allowed_retries = 2
        self.current_retries = 0
        self.pod_template_map = {}

        super().__init__(*args,host=host, port=port, ssl=ssl,**kwargs)

    async def configure(self):
        self.incoming_queue = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()
        await self.startup()
        
    async def startup(self):
        try:
            await self.connect_websocket()
        except:
            self.current_retries += 1
            if self.current_retries < self.allowed_retries:
                sleeping_time = (self.current_retries + 1)
                logger.error(f"Initial Connection failing: Trying again in {sleeping_time} seconds")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                return

        consumer_task = asyncio.create_task(
            self.consumer()
        )

        worker_task = asyncio.create_task(
            self.workers()
        )



        done, pending = await asyncio.wait(
            [consumer_task, worker_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        logger.error(f"Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Reconnecting')

        for task in pending:
            task.cancel()

        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup() # Attempt to ronnect again
        

    async def connect_websocket(self):

        hostable_pods = [str(podid) for podid, function in self.podid_pod_map.items()]
        hostable_pods_qstring = ",".join(hostable_pods)

        uri = f"{self.websocket_protocol}://{self.websocket_host}:{self.websocket_port}/peasent/{self.unique_name}/?token={self.token}&pods={hostable_pods_qstring}"
        
        self.connection = await websockets.client.connect(uri)
        # Our initial payload will be the Pods that were registered! Our allowance
        message = await self.connection.recv()
        allowance = AllowanceMessage.from_channels(message=message)

        self.pod_template_map = allowance.data.pod_template_map
        logger.info(f"We are able to host these pods {[ pod for pod, template in self.pod_template_map.items()]}")


    async def consumer(self):
        logger.warning(" [x] Awaiting Node Calls")
        async for message in self.connection:
            await self.incoming_queue.put(message)


    async def send_to_connection(self, message: AssignationMessage):
        await self.connection.send(message.to_channels())
       

    async def workers(self):
        while True:
            message = await self.incoming_queue.get()
            message = AssignationMessage.from_channels(message)
            assert message.data.pod is not None, "Received assignation that had no Pod?"
            asyncio.create_task(self.podid_function_map[message.data.pod](message)) # Run in parallel
            self.incoming_queue.task_done()


    




from bergen.auths.legacy.app import LegacyApplication
from bergen.auths.backend.app import ArnheimBackendOauth
from bergen.auths.implicit.app import ImplicitApplication
from bergen.clients.base import BaseBergen
from bergen.enums import GrantType
import os

class Bergen(BaseBergen):

    def __init__(self, 
    host: str = None, 
    port: int = None,
    client_id: str = None, 
    client_secret: str = None,
    grant_type: GrantType = GrantType.BACKEND,
    protocol="http", bind=True,
    allow_insecure=None,
    is_local=None,
    **kwargs) -> None:

        
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1" if allow_insecure is not None else os.getenv("OAUTHLIB_INSECURE_TRANSPORT", "0")
        os.environ["ARNHEIM_LOCAL"] = "0" if is_local is not None else os.getenv("ARNHEIM_LOCAL", "0")

        protocol = protocol or os.getenv("ARNHEIM_PROTOCOL", "http")
        host = host or os.getenv("ARNHEIM_HOST", "localhost")
        port = int(port or os.getenv("ARNHEIM_PORT", "8000"))
        client_id = client_id or os.getenv("ARNHEIM_CLIENT_ID", None)
        client_secret = client_secret or os.getenv("ARNHEIM_CLIENT_SECRET", None)
        grant_type = grant_type or os.getenv("ARNHEIM_GRANT_TYPE", GrantType.BACKEND)
        

        if grant_type == GrantType.BACKEND: auth = ArnheimBackendOauth(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        elif grant_type == GrantType.IMPLICIT: auth = ImplicitApplication(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        elif grant_type == GrantType.PASSWORD: auth = LegacyApplication(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        else: raise NotImplementedError("Please Specifiy a valid Grant Type")

        super().__init__(auth=auth, host=host, port=port, protocol=protocol, auto_negotiate=True, bind=bind, **kwargs)


from bergen.messages.types import ALLOWANCE
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import List, Mapping, Optional


class AllowanceMetaModel(MessageMetaModel):
    type: str = ALLOWANCE

class AllowanceDataModel(MessageDataModel):
    pod_template_map: Mapping[int, int]


class AllowanceMessage(MessageModel):
    data: AllowanceDataModel
    meta: AllowanceMetaModel = {"type": ALLOWANCE}



from abc import ABC, abstractmethod
from bergen.enums import ClientType


class AuthError(Exception):
    pass


class BaseAuthBackend(ABC):



    def __init__(self) -> None:
        super().__init__()


    @abstractmethod
    def getToken(self, loop=None) -> str:
        raise NotImplementedError("This is an abstract Class")


    @abstractmethod
    def getClientType(self) -> ClientType:
        raise NotImplementedError("This is an Abstract Class")

    @abstractmethod
    def getProtocol(self):
        raise NotImplementedError("This is an Abstract Class")


from bergen.messages.types import ASSIGNATION
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import Optional


class AssignationParams(BaseModel):
    pass

class AssignationMetaExtensionsModel(MessageMetaExtensionsModel):
    progress: Optional[str]

class AssignationMetaModel(MessageMetaModel):
    type: str = ASSIGNATION
    extensions: Optional[AssignationMetaExtensionsModel]

class AssignationDataModel(MessageDataModel):
    id: int
    node: Optional[int] #TODO: Maybe not optional
    pod: Optional[int]
    template: Optional[int]
    status: Optional[str]
    statusmessage: Optional[str]
    reference: str
    callback: Optional[str]
    progress: Optional[str]

    inputs: dict
    outputs: Optional[dict]
    params: Optional[AssignationParams]


class AssignationMessage(MessageModel):
    data: AssignationDataModel
    meta: AssignationMetaModel


from typing import Any
from bergen.schema import AssignationParams, Pod, PodStatus, ProvisionParams
from bergen.registries.arnheim import get_current_arnheim
import asyncio
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
import logging
from aiostream import stream
import uuid
from bergen.extenders.contexts.pod import HostedPod

class NodeExtender(BaseExtender):


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        bergen = get_current_arnheim()

        self._postman = bergen.getPostman()
        self._loop, self._force_sync = bergen.getLoopAndContext()


    def provide(self, params: ProvisionParams, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.provide_async(params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.provide_async(params, **kwargs))
            return result

    async def provide_async(self, params: ProvisionParams, **kwargs):
        return await self._postman.provide(self, params, **kwargs)       

    async def assign_async(self, inputs: dict, params: AssignationParams, **kwargs):
        
        return await self._postman.assign(self, inputs, params, **kwargs)

    async def delay_async(self, inputs: dict, params: AssignationParams, **kwargs):
    
        return await self._postman.delay(self, inputs, params, **kwargs)

    def stream(self, inputs: dict, params: AssignationParams = None, **kwargs):

        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs))
    
    def delay(self, inputs: dict, params: AssignationParams = None, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.delay_async(inputs, params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.delay_async(inputs, params, **kwargs))
            return result


    def __call__(self, inputs: dict, params: AssignationParams = None, **kwargs) -> dict:
        """Call this node (can be run both asynchronously and syncrhounsly)

        Args:
            inputs (dict): The inputs for this Node
            params (AssignationParams, optional): [description]. Defaults to None.

        Returns:
            outputs (dict): The ooutputs of this Node
        """
        if self._loop.is_running() and not self._force_sync:
            return self.assign_async(inputs, params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.assign_async(inputs, params, **kwargs))
            return result



    def hosted(self, params: ProvisionParams = {}, enter_when = PodStatus.ACTIVE, **kwargs):
        return HostedPod(self, params = {}, enter_when=enter_when, **kwargs)

    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from typing import Optional
from pydantic.fields import Field
from pydantic.main import ModelMetaclass, BaseModel



class ArnheimObject(BaseModel, metaclass=ModelMetaclass):
    TYPENAME: str = Field(None, alias='__typename')

    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(self.__dict__, indent=4, width=1)

    def _repr_html_(self):
        
        def buildTable(attributes):
            tablestring = "<table>"
            for key, value in attributes.items():
                tablestring = tablestring + (f"""
                    <tr>
                        <td>{key.capitalize()}</td>
                        <td>{value}</td>
                    </tr>
                """)
            return tablestring + "</table>"



        return f"""
            <p> Instance of {self.__class__.__name__} <p>
            {buildTable(self.__dict__)}
        """